#!/usr/bin/perl -w

# execution.pl -- A CGI/Perl script.

#                 Receives a GET with a fasta file an execution type ("Parametric" or "Sequential")
#                 and calls a BLAST execution through the InteGrade's API.
#
#                 It returns a page with either the execution ID or an error message, and a web
#                 link to return to the submission frame.

# M.S.Reis, 25.03.2012

use strict;
use CGI;

# Base64
use MIME::Base64;

# Allows XML POSTs
require LWP::UserAgent;

#----- parameters -----

# BLAST Application ID (the BLAST application and its binary must be already available at the
# InteGrade's Application Repository!)
#
my $APPLICATION_ID = 1;

# Complete path of the database (TODO: uses NFS to do it)
#
my $DATABASE = '/usr/local/blastdb/nr';

# Program: blastx: translated nucleotide against protein
#          blastn: nucleotide against nucleotide
#          blastp: protein against protein
my $PROGRAM = 'blastx';

# Input HTTP address
#
my $INPUT_ADDRESS = 'http://TestPageServer.com/cgi-bin/blast/execution.pl';

# Sample HTTP address
#
my $SAMPLE_ADDRESS = 'http://TestPageServer.com/blast/lepto.fasta';

# Address of the execution submission
#
my $SUBMISSION_ADDRESS = 'http://InteGradeAPI.com:9090/executions';

# Maximum number of available nodes in the InteGrade middleware; important information
# for a parametric execution
#
my $MAX_NUMBER_OF_NODES = 6;

# Absolute path of the data directory
#
my $DATA_PATH = "/var/www/blast/data";

# List of Execution IDs file
#
my $EXEC_ID_FILE = "exec_ids.in";

#----- main program -----

# if the method is GET, it prints the HTML form
# and ends the program
#
if ($ENV{REQUEST_METHOD} eq 'GET'){
  print "Content-type: text/html\n\n";
  print <<"HTML";
<HTML>
<HEAD>
<TITLE>Input frame</TITLE>
</HEAD>
<H3>Sequence alignment using the BLAST tool running on InteGrade middleware</H3>
<P>
This test consists in a comparison of the nucleotide target sequence(s) against the GenBank's
protein database (nr). To perform the test please follow these three steps.
<form name="input" action="$INPUT_ADDRESS" method="post">
<UL>
  <LI> Paste your target sequences data (FASTA format) into the box below (you may take them either from
  <a href="$SAMPLE_ADDRESS" target="_blank">this sample file</a> or directly from 
  <a href="ftp://ftp.ncbi.nih.gov/genomes/" target="_blank">GenBank</a>):
  <P>
  <textarea rows="8" cols="80" name="sequences"></textarea></LI>
  <LI> Select the execution type:<br />
  <input type="radio" name="type" value="sequential" checked /> Sequential (the sequences will be processed by a single node)<br />
  <input type="radio" name="type" value="parametric" /> Parametric (the sequences will be split into up to $MAX_NUMBER_OF_NODES files and they will be processed by different nodes)
  <BR /><BR /></LI>
  <LI>And start the execution! <input type="submit" value="Submit" /></LI>
</UL>
<P>
The results will be showed in the frame below as soon as the execution is finished. 
</form>
HTML

  # end of CGI
  exit 0;
}


# reading data from the POST method
#
my $buffer;
if ($ENV{REQUEST_METHOD} eq 'POST'){
    read(STDIN, $buffer, $ENV{CONTENT_LENGTH});
}

# parsing the application type and the fasta file
#
my ($type, $fasta_string);
foreach my $pair (split(/&/, $buffer)){
    my ($field, $value) = split("=", $pair);
    $value =~ s/\+//g;
    $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
    if ($field eq "type"){
      $type = $value;
    }    
    if ($field eq "sequences"){
      $fasta_string = $value;
    }    
}

my $post_data = "";
my $number_of_copies = 1; # default number of copies for a sequential execution

# if the application is sequential, creates a POST with sequential application type
#
if ($type eq "sequential"){

  $post_data = "
<execution>
  <applicationId>" . $APPLICATION_ID . "</applicationId>
  <applicationType>Sequential</applicationType>
  <constraints></constraints>
  <preferences></preferences>
  <arguments>-p $PROGRAM -o genes.$PROGRAM -i genes.fasta -d " . $DATABASE . " </arguments>
  <storeStdout> False </storeStdout>
  <storeStderr> False </storeStderr>
  <forceDifferentNodes> False </forceDifferentNodes>
  <numberOfTasks> 0 </numberOfTasks>
  <outputFiles>
      <executionFile>
          <fileName>genes.$PROGRAM</fileName>
      </executionFile>
  </outputFiles>
  <inputFiles>
      <executionFile>
          <fileName>genes.fasta</fileName>
          <file>" . encode_base64($fasta_string) . "</file>
      </executionFile>
  </inputFiles>
</execution>\n";

}

# if the application is parametric, splits the fasta file into
# $MAX_NUMBER_OF_NODES pieces and creates a POST with parametric application type
#
elsif ($type eq "parametric"){
  my @genes = split (/>/, $fasta_string); # split produces number_of_genes + 1 elements
  my @fasta_files;
  for (my $j = 0; $j < $MAX_NUMBER_OF_NODES; $j++){
    $fasta_files[$j] = "";
  }
  my $j = 0;
  for (my $i = 1; $i <= $#genes; $i++){
    $fasta_files[$j] = $fasta_files[$j] . ">" . $genes[$i];
    $j++;
    $j == $MAX_NUMBER_OF_NODES and $j = 0;
  }

  # the number of copies of a parametric execution is the minimum between the maximum
  # number of nodes available at InteGrade and the number of sequences that are contained in the FASTA file.
  #
  $number_of_copies = $#genes;
  $MAX_NUMBER_OF_NODES < $number_of_copies and $number_of_copies = $MAX_NUMBER_OF_NODES;

  $post_data = "
<execution>
  <applicationId>" . $APPLICATION_ID . "</applicationId>
  <applicationType>Parametric</applicationType>
  <numberOfCopies>" . $number_of_copies . "</numberOfCopies>
  <forceDifferentNodes> True </forceDifferentNodes>
  <storeStdout> False </storeStdout>
  <storeStderr> False </storeStderr>
  <parametricCopies>";

  for (my $j = 1; $j <= $number_of_copies; $j++){
    $post_data = $post_data . "
      <parametricCopy>
          <inputFileNames>
                <executionFile><fileName>gene" . $j . ".fasta</fileName></executionFile>
          </inputFileNames>
          <outputFileNames>
                <executionFile><fileName>gene" . $j . ".$PROGRAM</fileName></executionFile>
          </outputFileNames>
          <arguments>-p $PROGRAM -i gene" . $j . ".fasta -o gene" . $j . ".$PROGRAM -d " . $DATABASE . "</arguments>
      </parametricCopy>";
  }
  
  $post_data = $post_data . "
  </parametricCopies>
  <inputFiles>";

  for (my $j = 1; $j <= $number_of_copies; $j++){
    $post_data = $post_data . "
      <executionFile>
          <fileName>gene" . $j .  ".fasta</fileName>
          <file>" . encode_base64($fasta_files[$j-1]) . "</file>
      </executionFile>";
  }

  $post_data = $post_data . "
  </inputFiles>
</execution>";
}

else{
  print "Content-type: text/html\n\n";
  die "Error: unknown application type!\n";
}

# requesting an execution through the InteGrade's API
#
my $exec_id = -1;
my $userAgent = LWP::UserAgent->new;
$userAgent->timeout(30);
$userAgent->env_proxy;


my $response = $userAgent->post($SUBMISSION_ADDRESS, Content_Type => 'text/xml', Content => $post_data);
$response->is_success or die $response->status_line;

if ($response->as_string =~ /\d+\/execution\/(\d+)/){
  # parses the "Location" field that contains the execution ID
  $exec_id = $1;   
  
  # writing the new execution IDs data file
  #
  open (ARQ, ">>$DATA_PATH/$EXEC_ID_FILE") or die "Could not open $EXEC_ID_FILE!\n";
  printf ARQ "%d %d\n", $exec_id, $number_of_copies;
  close (ARQ);  
}

# creating the HTML document with the obtained execution ID
#
print "Content-type: text/html\n\n";
print <<"HTML";
<HTML>
  <HEAD>
    <TITLE>Input frame</TITLE>
  <H3>Sequence alignment using the BLAST tool running on InteGrade middleware</H3>
  <P><BR /><BR />
  <CENTER>
    Execution started successfully! The execution ID is: <B>$exec_id</B><BR />
    Please follow the execution status in the results frame<BR /><BR />
    <A HREF="$INPUT_ADDRESS">Back to the submission page</A>
  </CENTER>
</HTML>
HTML

# end of CGI
exit 0;
